GRIT
====

Lightweight, asynchronous EventEmitter- and [Backbone](http://documentcloud.github.com/backbone/)-inspired 
models for NodeJS and browsers. For when the whole of Backbone seems too hefty, or synchronicity is 
hurting your app's response times.

Sample
------
In pure JS:

```javascript
var User = Grit.extend(function(username) {
  Grit.call(this, {username:username});
});

User.require('username');

User.validate('username', function(username) {
  return typeof username === 'string';
});

var user = new User('reissbaker');
user.on('change:username', function(newUsername) {
  alert('changing username from ' + user.get('username') + ' to ' + newUsername);
});
user.set('username', 'rekabssier');
```

Or, in CoffeeScript:

    class User extends Grit
      @required 'username'
      @validate 'username', (username) ->
        typeof username == string
  
      constructor: (username) ->
        super username: username

    user = new User 'reissbaker'
    user.on 'change:username', (oldUsername, newUsername) ->
      alert "changing username from #{oldUsername} to #{newUsername}"
    user.set 'username', 'rekabssier'

API
---

The API follows the NodeJS EventEmitter API, and the events emitted are
inspired by Backbone.

### Constructor
* `new Grit(properties)` creates a new Grit object, and will set the
properties according to the properties hash passed in. Please note
that if you mark properties as required with `Grit.required`, you _must_ 
either pass their initial values into the constructor or set initial
values for them with `Grit.defaults`.

### Instance Methods
* `grit.on(evt, handler)`, or `grit.addListener(evt, handler)` add an event
handler to the object. Grit objects emit `"change:p"` events, where `p`
is the name of the property being changed.
* `grit.removeListener(e, handler)` removes an event handler.
* `grit.get(property, callback)` gets a named property and passes it to the
callback. `callback` should be of the form `function(err, data)`, like most
NodeJS callbacks.
* `grit.set(property, value[, silent, callback])` sets a property. Passing a truthy
value as the `silent` parameter suppresses a change event; otherwise, a
`change:property` event will be emitted, where `property` is the value
passed in as the `property` parameter. The callback will be called once
the set has propagated to the server or DB through the `Grit.update` method.
Callbacks should be of the form `function(err)`.
* `grit.destroy(callback)` destroys the model on the server or DB by proxying
out to `Grit.destroy`. The callback will be called when the destruction has
propagated to the server or DB, and should be of the form `function(err)`.

### Class Methods
#### Syncing
* `Grit.create(model, data, callback)` is the method to override for
syncing model creation. `model` is the model itself, and `data` is the
property data in the raw. `callback` is a function of the form
`function(err)`, and should be called once the server or DB is updated.
* `Grit.read(model, data, property, callback` is the method to override
for reading individual properties from the server or DB. `model` is the
model being read from, `data` is the model's data in the raw (you probably
don't need to touch it!), `property` is the property currently being read,
and `callback` is a callback of the form `function(err, data)`, and you
should pass successfully-retrieved data to it (or call it with an error).
* `Grit.update(model, data, property, callback)` is the method to override
for updating individual properties on the server. `model` is the model
currently being updated, `data` is the model's data in the raw (grab the
property data from there, rather than asynchronously trying to get it with
get), `property` is the property currently in need of syncing, and `callback`
is a function of the form `function(err)`, which you should call (without
an error) on successful completion, or with an error upon an error.
* `Grit.destroy(model, data, callback)` is the method to override for
syncing model destruction. `model` is the model currently being destroyed,
`data` is its property data in the raw, and `callback` is a function of
the form `function(err)`, which should be called once the server or DB
is updated.

#### Property defaults, assertions
* `Grit.defaults` accepts a hash of default values to set for properties
on model construction. If you don't specify default values for required
properties, you _must_ pass them into the constructor.
* `Grit.required(property)` marks a property as required. All required
properties that do not have default values set MUST be passed at 
initialization into the Grit constructor. If a required property is 
missing, all calls to `set` will fail to validate.
* `Grit.unrequired(property)` marks a formerly-required property as
no longer required.
* `Grit.validate(property, validator)` adds a validation callback to
a property. Validation callbacks must be of the form `function(value)`,
and the `value` that gets passed in is the new value being set. If
all validators of a property return truthy values, the property will 
successfully be set; otherwise, it won't.

#### Inheritance
* `Grit.extend(constructorFunction)` makes it easy for normal JS classes
to inherit from Grit. Just pass the constructor 
function to `extend`, and don't forget to call `Grit.call(this)` (or
`SubClass.call(this)`) on the first line of the constructor function.
If you're using CoffeeScript, just use the native `extends` keyword,
and if you override the constructor, don't forget to call `super` before
trying to access any Grit methods.

LICENSE
=======

GPL v3. Check LICENSE.txt and "GNU General Public License".txt for full details.
