(function() {
	var Grit, EventEmitter;

	/*
	 * require in NodeJS's native EventEmitter class,
	 * or emulate it in the browser.
	 */

	if(typeof exports !== 'undefined') {
		EventEmitter = require('events').EventEmitter;
	} else {
		EventEmitter = function() {
			var handlers = {};

			this.on = this.addListener = function(e, callback) {
				if(!handlers.hasOwnProperty(e)) {
					handlers[e] = [];
				}
				handlers[e].push(callback);
			};

			this.removeListener = function(e, callback) {
				var handleList, index, length, curr;
				if(handlers.hasOwnProperty(e)) {
					handleList = handlers[e];
					for(index = 0, length = handleList.length; index < length; index++) {
						curr = handleList[index];
						if(curr === callback) {
							handleList.splice(index, 1);
							if(handleList.length === 0)
								delete handlers[e];
							return;
						}	
					}
				}
			};

			this.emit = function(e) {
				var handleList, index, length, args;
				if(handlers.hasOwnProperty(e)) {
					handleList = handlers[e];
					args = Array.prototype.slice(this, arguments).slice(1);
					for(index = 0, length = handleList.length; index < length; index++)
						handleList[index].call(this, args);
				}
			};
		};
	}

	/*
	 * The model class
	 * ===============
	 */

	Grit = function() {
		var data, prop, valid, changeData, 
				validateData, ensurePresenceOfData,
				properties, isNew, finishedCreating,
				args;

		// call super
		EventEmitter.call(this);

		// get default arguments
		args = Array.prototype.slice.call(arguments, 0);
		if(args.length === 1) {
			if(typeof args[0] === 'function') {
				properties = {};
				finishedCreating = args[0];
				isNew = true;
			} else if(typeof args[0] === 'boolean') {
				properties = {};
				finishedCreating = function(){};
				isNew = args[0];
			} else {
				properties = args[0];
				finishedCreating = function(){};
				isNew = true;
			}
		} else if(args.length === 2) {
			if(typeof args[0] === 'boolean') {
				properties = {};
				isNew = args[0];
				finishedCreating = args[1];
			} else  {
				properties = args[0];
				if(typeof args[1] === 'boolean') {
					isNew = args[1];
					finishedCreating = function() {};
				} else {
					isNew = true;
					finishedCreating = args[1];
				}
			}
		} else {
			properties = args[0];
			isNew = args[1];
			finishedCreating = args[2]
		}
		
		// initialise data
		data = {};

		/* 
		 * helper functions
		 * ----------------
		 */

		// checks if all `required` data is present
		ensurePresenceOfData = function(that) {
			var prop;
			if(!that.constructor.hasOwnProperty('reqs') || !that.constructor.reqs)
				return true;

			for(prop in that.constructor.reqs) {
				if(that.constructor.reqs.hasOwnProperty(prop) && !data.hasOwnProperty(prop))
					return false;
			}
			return true;
		};

		// checks if the value is valid for the property
		validateData = function(that, name, value) {
			var validators, valid, index, length;
			if(!that.constructor.hasOwnProperty('validations') || !that.constructor.validations)
				return true;
			if(!that.constructor.validations.hasOwnProperty(name) || !that.constructor.validations[name])
				return true;
			else {
				validators = that.constructor.validations[name];
				for(index = 0, length = validators.length; index < length; index++) {
					if(!validators[index](value))
						return false;
				}
			}
			return true;
		};

		// changes the data, emitting the appropriate events and calling the callbacks
		changeData = function(that, name, value, silent, finished) {
			if(typeof silent === 'undefined' || !silent) {
				that.get(name, function(err, oldValue) {
					if(!isNew)
						that.emit('change:' + name, that, oldValue, value);
					data[name] = value;
					if(!isNew)
						that.constructor.update(that, data, name, finished);
				})
			} else {
				data[name] = value;
				if(!isNew)
					that.constructor.update(that, data, name, finished);
			}
		};

		/* 
		 * priveleged instance methods
		 * ---------------------------
		 */

		this.get = function(name, callback) {
			if(data.hasOwnProperty(name))
				callback(null, data[name]);
			else {
				this.constructor.read(this, data, name, function(err, value) {
					if(!err)
						data[name] = value;
					callback(err, value);
				});	
			}
		};

		this.set = function(name, value, silent, finished) {
			var name, value, silent, finished, valid, args;

			args = Array.prototype.slice.call(arguments, 0);
			name = args[0];
			value = args[1];
			if(args.length === 2) {
				silent = false;
				finished = function(){};
			} else if(args.length === 3) {
				if(typeof args[2] === 'function') {
					silent = false;
					finished = args[2];
				} else {
					silent = args[2];
					finished = function(){};
				}
			} else {
				silent = args[2];
				finished = args[3];
			}

			valid = validateData(this, name, value) && ensurePresenceOfData(this);

			if(valid)
				changeData(this, name, value, silent, finished);
			else
				finished('Error: ' + value + ' is an invalid value for ' + name + ' of ' + this);
		};

		this.destroy = function(callback) {
			this.constructor.destroy(this, data, callback);
		};

		/*
		 * setting the initial data from the defaults and
		 * the properties passed in
		 * ------------------------
		 */

		// if there aren't properties, set them to empty
		if(typeof properties === 'undefined')
			properties = {};

		if(this.constructor.hasOwnProperty('defaults')) {
			for(prop in this.constructor.defaults) {
				if(this.constructor.defaults.hasOwnProperty(prop) && !properties.hasOwnProperty(prop))
					properties[prop] = this.constructor.defaults[prop];
			}
		}

		// check validity
		valid = true;
		if(this.constructor.hasOwnProperty('reqs')) {
			for(prop in this.constructor.reqs) {
				if(!properties.hasOwnProperty(prop) || !validateData(this, prop, properties[prop])) {
					valid = false;
					break;
				}
			}
		}	

		// set the defaults for things not being overridden; otherwise, set the properties
		if(valid)
			for(prop in properties)
				if(properties.hasOwnProperty(prop))
					changeData(this, prop, properties[prop], isNew, function(){});

		this.constructor.create(this, data, finishedCreating);
		isNew = false;
	};

	/* 
	 * inheritance
	 * ===========
	 */

	Grit.prototype = new EventEmitter();
	Grit.prototype.constructor = Grit;

	Grit.extend = function(structFn) {
		structFn.prototype = new Grit();
		structFn.prototype.constructor = structFn;
		for(var prop in Grit) {
			if(Grit.hasOwnProperty(prop))
				structFn[prop] = Grit[prop];
		}
		return structFn;
	};

	/*
	 * sync functions
	 * ==============
	 */

	Grit.create = function(model, data, callback) {
		if(typeof callback !== 'undefined')
			callback(null);
	};
	Grit.read = function(model, data, property, callback) {
		callback('Error: no ' + property + ' property could be read', null);
	};
	Grit.update = function(model, data, property, callback) {
		callback(null);
	};
	Grit.destroy = function(model, data, callback) {
		callback(null);
	};
	Grit.find = function(callback) {
		callback('Error: find must be overridden.', null);
	};

	/*
	 * Property defaults, assertions
	 * =============================
	 */

	Grit.defaults = function(properties) {
		if(!this.hasOwnProperty('defaults'))
			this.defaults = {};
		for(var prop in properties)
			if(properties.hasOwnProperty(prop))
				this.defaults[prop] = properties[prop];
	};

	Grit.required = function() {
		var args, index; 
		args = Array.prototype.slice.call(arguments, 0);

		if(!this.hasOwnProperty('reqs'))
			this.reqs = {};

		for(index = 0; index < args.length; index++)
			this.reqs[args[index]] = true;
	};
	Grit.unrequired = function() {
		var args, index;
		args = Array.prototype.slice.call(arguments, 0);

		if(!this.hasOwnProperty('reqs'))
			return;

		for(index = 0; index < args.length; index++) {
			if(this.reqs.hasOwnProperty(args[index]))
				delete this.reqs[args[index]];
		}
	};

	Grit.validate = function(name, validator) {
		if(!this.hasOwnProperty('validations'))
			this.validations = {};

		if(!this.validations.hasOwnProperty(name))
			this.validations[name] = [];
		this.validations[name].push(validator);
	};
	Grit.removeValidation = function(name, validator) {
		var validators, index, length;
		if(!this.hasOwnProperty('validations'))
			return;

		if(this.validations.hasOwnProperty(name)) {
			validators = this.validations[name];
			for(index = 0, length = validators.length; index < length; index++) {
				if(validators[index] === validator) {
					validators.splice(index, 1);
					break;
				}
			}

			if(validators.length === 0)
				delete this.validations[name];
		}
	};

	/* 
	 * Export
	 * ======
	 */

	if(typeof exports !== 'undefined') {
		exports = module.exports = Grit;
	} else {
		window.Grit = Grit;
	}
}());
